import logging
from logging.config import dictConfig
from typing import Any, Callable, Dict, List

import structlog
from structlog.contextvars import merge_contextvars

from .utils import CleanseLogArgumentsFilter, cleanse_dict_processor

try:
    from structlog_sentry import SentryProcessor
except ImportError:
    has_sentry = False
else:
    has_sentry = True


class BaseLoggingConfig:
    #: This processor is used as long as ``debug`` is ``False``. It's the one
    #: intended for production and testing use.
    default_processor = structlog.processors.JSONRenderer()

    #: When debugging an application it can be worth changing the default
    #: processor to something more readable. This processor is used when
    # ``debug`` is set to ``True``.
    debug_processor = structlog.dev.ConsoleRenderer()

    #: Each application may have its own required processors. These processors
    #: can be added by setting this list.
    additional_processors: List[Callable[[str, str, Dict], Dict]] = []

    #: During debugging one might want to change the formatting or the handling
    #: of event data, for example, pretty-print exception tracebacks.
    debug_processors: List[Callable[[str, str, Dict], Dict]] = [
        structlog.processors.format_exc_info  # type: ignore
    ]

    #: Some loggers are fairly noisy and often not useful to us. Adding
    #: additional loggers to this list to send them to a blackhole.
    default_blackhole_loggers = ["urllib3"]

    #: Similarly to the ``default_blackhole_loggers``, during testing we might
    #: want to send additional loggers to a black hole. Add these loggers here.
    testing_blackhole_loggers = ["", "flake8"]

    @classmethod
    def configure(
        cls, *, level: str = "WARNING", testing: bool = False, debug: bool = False
    ) -> None:
        # Configure logging according to structlog's suggestion:
        # https://www.structlog.org/en/stable/standard-library.html#rendering-using-structlog-based-formatters-within-logging
        shared_processors = [
            merge_contextvars,
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            structlog.processors.TimeStamper(fmt="ISO"),
        ]

        # Construct the list of loggers we want to configure explicitly
        loggers: Dict[str, Dict[str, Any]] = {
            "": {"handlers": ["default"], "level": level, "propagate": True}
        }
        for logger in cls.default_blackhole_loggers:
            loggers[logger] = {"handlers": ["blackhole"], "propagate": False}
        if testing:
            for logger in cls.testing_blackhole_loggers:
                loggers[logger] = {"handlers": ["blackhole"], "propagate": False}

        # Construct the list of processors and the final processor that prints
        # or outputs the log event at the end.
        processor = cls.default_processor
        additional_processors = []
        if has_sentry:
            # Only include when installed.
            # Submit the log message to Sentry in a way that allows
            # them to perform grouping on duplicate events.
            # Refs: https://github.com/getsentry/sentry-python/issues/228
            additional_processors.append(
                SentryProcessor(level=logging.INFO, event_level=logging.ERROR)
            )
        additional_processors.extend(cls.additional_processors)
        if debug:
            processor = cls.debug_processor  # type: ignore
            additional_processors.extend(cls.debug_processors)

        dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "filters": {
                    "cleanse": {"()": CleanseLogArgumentsFilter}  # for stdlib loggers
                },
                "formatters": {
                    "json": {
                        "()": structlog.stdlib.ProcessorFormatter,
                        "processors": [
                            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                            processor,
                        ],
                        "foreign_pre_chain": (
                            shared_processors + [cleanse_dict_processor]
                        ),
                    }
                },
                "handlers": {
                    "default": {
                        "level": "DEBUG",
                        "class": "logging.StreamHandler",
                        "filters": ["cleanse"],
                        "formatter": "json",
                    },
                    "blackhole": {"class": "logging.NullHandler"},
                },
                "loggers": loggers,
            }
        )

        structlog.configure(
            processors=(
                shared_processors  # type: ignore
                + [
                    structlog.stdlib.PositionalArgumentsFormatter(),
                    structlog.processors.StackInfoRenderer(),
                    *additional_processors,
                    cleanse_dict_processor,
                    structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
                ]
            ),
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=not testing,
        )
