import logging
import re
from collections.abc import Mapping

_SENSITIVE_KEYS = ["CONN", "DSN", "KEY", "PASS", "SECRET", "TOKEN"]
SENSITIVE_KEYS = re.compile(
    "|".join(map(re.escape, _SENSITIVE_KEYS)), flags=re.IGNORECASE
)
CLEANSED_SUBSTITUTE = "****************"


def cleanse_dict(d):
    """
    Cleanse a key/value of sensitive content. If the value is a dictionary,
    recursively cleanse the keys in that dictionary.

    Adopted from `Django's cleanse_settings()
    <https://github.com/django/django/blob/3e7c39f710ec1895611fa1b337423742eebd0da1/django/views/debug.py#L83-L106>`_.
    """

    def _cleanse(key, value):
        try:
            if SENSITIVE_KEYS.search(key):
                return CLEANSED_SUBSTITUTE
            elif isinstance(value, Mapping):
                return cleanse_dict(value)
            elif isinstance(value, (list, tuple)):
                return [_cleanse("", v) for v in value]
            else:
                return value
        except TypeError:
            # If the key isn't regex-able, just return as-is.
            return value

    return {key: _cleanse(key, value) for key, value in d.items()}


class CleanseLogArgumentsFilter(logging.Filter):
    """
    This is a filter which ensures that all dictionaries passed as arguments to
    a record are cleaned from secrets.
    """

    def filter(self, record):
        if isinstance(record.args, tuple):
            # multiple arguments given
            record.args = tuple(
                (cleanse_dict(arg) if isinstance(arg, dict) else arg)
                for arg in record.args
            )
        elif isinstance(record.args, dict):
            # One can pass a single dictionary to use named string interpolation such
            # as `"some string '%(x)s' here" % {"x": "abc"}`. In that case we want to
            # still clean each value accordingly.
            record.args = cleanse_dict(record.args)
        return True


def cleanse_dict_processor(logger, method_name, event_dict):
    return cleanse_dict(event_dict)
