import re
from distutils.version import Version
from typing import Optional, Tuple


class CrateVersion(Version):
    """Version numbering for CrateDB releases.

    Implements the standard interface for version number classes as
    described above. A CrateDB version consists of two or three
    dot-separated numeric components, with an optional "-SNAPSHOT" tag
    on the end. Alternatively, a nightly version consists of the term "nightly"
    optionally followed by three dot-separated numeric components indicating
    the version, as well as an optional datetime or date component for the day of the
    build in the form ``YYYY-MM-DD-HH-mm`` or ``YYYYMMDD.

    The following are valid version numbers (shown in the order that
    would be obtained by sorting according to the supplied cmp function):

    - 0.1
    - 3.2
    - 3.2.0
    - nightly-3.2.1
    - nightly-3.2.1-20190708
    - nightly-3.2.1-2019-07-08-01-02
    - 3.2.1-SNAPSHOT
    - 3.2.1
    - 4.0.0-SNAPSHOT
    - nightly

    The following are examples of invalid version numbers:

    - 1
    - 1.2-SNAPSHOT
    - nightly-3.2
    - nightly-3.2.1-201907080200
    - nightly-3.2.1-201907080200-abc9876

    The rationale for this version numbering system will be explained
    in the distutils documentation.
    """

    version: Optional[Tuple[int, int, int]]

    major: int
    minor: int
    hotfix: int

    stable: bool
    snapshot: bool
    nightly: bool

    date: Optional[str] = None
    date_normalized: Optional[str] = None

    #: The regular expression to match a version string. Valid values are e.g.:
    #:
    #: - ``0.1``
    #: - ``0.1.2``
    #: - ``0.1.2-SNAPSHOT``
    #:
    #: Invalid values are, e.g:
    #:
    #: - ``0``
    #: - ``0.1-SNAPSHOT``
    version_re = re.compile(
        r"""
            ^  # Start of the string
            (?P<major>\d+)\.(?P<minor>\d+)  # Every version has a major.minor
            (\.(?P<hotfix>\d+)(?P<snapshot>-SNAPSHOT)?)?  # optionally hotfix
            $  # End of the string
        """,
        re.ASCII + re.VERBOSE,
    )

    #: The regular expression to match a nightly version string. Valid values
    #: are e.g.:
    #:
    #: - ``nightly``
    #: - ``nightly-0.1.2``
    #: - ``nightly-0.1.2-20190708``
    #: - ``nightly-0.1.2-2019-07-08-01-02``
    #:
    #: Invalid values are, e.g:
    #:
    #: - ``nightly-0.1``
    #: - ``nightly-0.1.2-201907080200``
    #: - ``nightly-0.1.2-201907080200-abc0987``

    nightly_version_re = re.compile(
        r"""
            ^
            (?:nightly)
            (?:
                -(?P<has_version>(?P<major>\d+)\.(?P<minor>\d+)\.(?P<hotfix>\d+))
                (?:
                    -(?P<date>\d{8}|\d{4}-\d{2}-\d{2}-\d{2}-\d{2})  # Match either yyyy-MM-dd-HH-mm or yyyymmdd
                )?
            )?$
        """,
        re.ASCII + re.VERBOSE,
    )

    def __init__(self, vstring: str):
        self.parse(vstring)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self}')"

    def __str__(self):
        if self.stable:
            return ".".join(map(str, self.version))

        if self.nightly:
            vstring = "nightly"
            if self.version:
                vstring += f"-{self.major}.{self.minor}.{self.hotfix}"
                if self.date:
                    vstring += f"-{self.date}"
            return vstring

        if self.snapshot:
            return f"{self.major}.{self.minor}.{self.hotfix}-SNAPSHOT"

    def _cmp(self, other):
        if isinstance(other, str):
            other = CrateVersion(other)

        if self.version and other.version and self.version != other.version:
            # For stable or snapshot releases, as well as for nightly versions
            # when both have version information, when the numeric versions
            # don't match the prerelease stuff doesn't matter.
            if self.version < other.version:
                return -1
            else:
                return 1

        if self.stable and other.stable:
            # 1.2.3 == 1.2.3
            return 0
        if self.snapshot and other.snapshot:
            # 1.2.3-SNAPSHOT == 1.2.3-SNAPSHOT
            return 0
        if self.nightly and other.nightly:
            if not self.version and not other.version:
                # nightly == nightly
                return 0
            if self.version and not other.version:
                # nightly-1.2.3 < nightly
                return -1
            if not self.version and other.version:
                # nightly > nightly-1.2.3
                return 1

            # nightly-$version ? nightly-$version
            assert self.version
            assert other.version

            if not self.date_normalized and not other.date_normalized:
                # nightly-1.2.3 == nightly-1.2.3
                return 0
            if self.date_normalized and not other.date_normalized:
                # nightly-1.2.3-20180708 < nightly-1.2.3
                return -1
            if not self.date_normalized and other.date_normalized:
                # nightly-1.2.3 > nightly-1.2.3-20180708
                return 1

            # nightly-1.2.3-$date ? nightly-1.2.3-$date
            assert self.date_normalized
            assert other.date_normalized

            if self.date_normalized < other.date_normalized:
                # 20190707 < 20190708
                return -1
            if self.date_normalized > other.date_normalized:
                # 20190708 > 20190707
                return 1

            # nightly-1.2.3-20190708 == nightly-1.2.3-20190708
            return 0
        if self.nightly and (other.snapshot or other.stable):
            if self.version:
                # nightly-1.2.3 < 1.2.3-SNAPSHOT | nightly-1.2.3 < 1.2.3
                return -1
            # nightly > 1.2.3-SNAPSHOT | nightly > 1.2.3
            return 1
        if self.snapshot and other.stable:
            # 1.2.3-SNAPSHOT < 1.2.3
            return -1
        if self.stable and (other.snapshot or other.nightly):
            if other.version:
                # 1.2.3 > 1.2.3-SNAPSHOT | 1.2.3 > nightly-1.2.3
                return 1
            # 1.2.3 < nightly
            return -1
        if self.snapshot and other.nightly:
            if other.version:
                # 1.2.3-SNAPSHOT > nightly-1.2.3
                return 1
            # 1.2.3-SNAPSHOT < nightly
            return -1

        assert False, "versions are not comparable"

    def __lshift__(self, other):
        assert isinstance(other, CrateVersion)

        if self.version and other.version:
            # nightly-1.2.3 << nightly-2.3.4
            return self.version < other.version
        # at least one does not have a version number, thus being a nightly w/o
        # version thus being greater than the other if the other does have a
        # version.
        if not self.version and not other.version:
            # nightly << nightly
            return False
        if self.version and not other.version:
            # nightly-1.2.3 << nightly
            return True

        # nightly << nightly-1.2.3
        return False

    def __rshift__(self, other):
        assert isinstance(other, CrateVersion)

        if self.version and other.version:
            # nightly-1.2.3 >> nightly-2.3.4
            return self.version > other.version
        # at least one does not have a version number, thus being a nightly w/o
        # version thus being greater than the other if the other does have a
        # version.
        if not self.version and not other.version:
            # nightly >> nightly
            return False
        if self.version and not other.version:
            # nightly-1.2.3 >> nightly
            return False

        # nightly >> nightly-1.2.3
        return True

    def parse(self, vstring):
        if vstring.startswith("nightly"):
            self._parse_nightly(vstring)
        else:
            self._parse_regular(vstring)

    def _parse_regular(self, vstring):
        match = self.version_re.match(vstring)
        if not match:
            raise ValueError("Invalid version number: '%s'" % vstring)

        major, minor, hotfix, snapshot = match.group(
            "major", "minor", "hotfix", "snapshot"
        )

        if hotfix:
            self.version = tuple(map(int, [major, minor, hotfix]))
        else:
            self.version = tuple(map(int, [major, minor])) + (0,)

        self.stable = False if snapshot else True
        self.nightly = False
        self.snapshot = True if snapshot else False
        self.major, self.minor, self.hotfix = self.version

    def _parse_nightly(self, vstring):
        match = self.nightly_version_re.match(vstring)
        if not match:
            raise ValueError("Invalid version number: '%s'" % vstring)

        has_version, major, minor, hotfix, date = match.group(
            "has_version", "major", "minor", "hotfix", "date"
        )

        if has_version:
            self.major, self.minor, self.hotfix = int(major), int(minor), int(hotfix)
            self.version = tuple(map(int, [self.major, self.minor, self.hotfix]))
        else:
            self.major, self.minor, self.hotfix = None, None, None
            self.version = None
        self.stable = False
        self.snapshot = False
        self.nightly = True

        self.date = date
        if self.date:
            self.date_normalized = date.replace('-', '')[:8]


CrateVersion("nightly-4.0.0")
